/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is gnugoclient, a program to connect go engines to go        *
 * servers. See http://www.lysator.liu.se/~gunnar/gnugoclient        *
 * for information.                                                  *
 *                                                                   *
 * Copyright 1998-2004 by Gunnar Farnebck.                          *
 * E-mail: gunnar@lysator.liu.se                                     *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License as    *
 * published by the Free Software Foundation - version 2             *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License in file COPYING for more details.      *
 *                                                                   *
 * You should have received a copy of the GNU General Public         *
 * License along with this program; if not, write to the Free        *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,           *
 * Boston, MA 02111, USA.                                            *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static Thread.Queue write_queue;
static object engine;
static int debug;
static function crash_callback = 0;

static mapping(int:function) callbacks = ([]);
static int id_number = 1;

static string dump_commands_file;
static string dump_responses_file;

// Write a message to the user. When running in screen, a CR seems to
// be required.
void trace(string s)
{
    werror(replace(s, "\n", "\r\n"));
}

// Send a command to the go engine. An optional callback can be
// specified, which is called with the response string when this
// arrives.
void send_command(string s, function|void callback)
{
    s = id_number + " " + s;
    if (callback)
	callbacks[id_number] = callback;
    if (debug)
	trace("Sent: " + s + "\n");
    write_queue->write(s + "\n");
    id_number++;
}

// Send a command to the go engine and wait for the response, which is
// returned.
//
// Warning: Never call this function from a callback initiated from
// send_command(). This will always deadlock since the same thread
// would need both to wait() and signal().
//
// Workaround: If you need to do this, use thread_create() to spin off
// a new thread which calls send_command_sync() and let the callback
// function return.
string send_command_sync(string s)
{
    string result;
    Thread.Condition cond = Thread.Condition();
    Thread.Mutex mutex = Thread.Mutex();
    object lock1;

    function callback(string s)
    {
	object lock2 = mutex->lock();
	result = s;
	cond->signal();
	destruct(lock2);
    };

    lock1 = mutex->lock();
    send_command(s, callback);
    cond->wait(lock1);
    destruct(lock1);
    return result;
}

void remove_callback(function callback)
{
    foreach (indices(callbacks), int n)
	if (callbacks[n] == callback)
	    m_delete(callbacks, n);
}

// Tell the program to stop playing. The shutdown() method makes a
// better job at cleaning up though.
void quit()
{
    crash_callback = 0;
    send_command_sync("quit");
}

static void program_trace_reader(object f)
{
    while (1)
    {
	string s = f->gets();
	if (!s)
	    break;
	if (debug)
	    trace("Stderr: " + s + "\n");
    }

    f->close();
    if (debug)
	trace("trace_reader closed down.\n");
}

static void program_reader(object f)
{
    int awaiting_new_response = 1;
    int id;
    string response;
    while (1)
    {
	string s = f->gets();
	if (!s)
	    break;
	if (debug)
	    trace("Recv: " + s + "\n");
	if (dump_responses_file)
	    Stdio.append_file(dump_responses_file, s + "\n");

	if (awaiting_new_response)
	{
	    if (s == "")
		continue;
	    if (s[0] != '=' && s[0] != '?')
		trace("Bad response: " + s + "\n");
	    else
	    {
		awaiting_new_response = 0;
		sscanf(s[1..], "%d %s", id, response);
	    }
	}
	else
	{
	    if (s != "")
		response += "\n" + s;
	    else
	    {
		if (callbacks[id])
		{
		    callbacks[id](response);
		    m_delete(callbacks, id);
		}
		awaiting_new_response = 1;
	    }
	}
    }

    f->close();
    if (debug)
	trace("Reader closed down.\n");
    if (crash_callback) {
	trace("Looks like a crash.\n");
	crash_callback();
    }
}

static void program_writer(object f)
{
    while (1)
    {
	string s = write_queue->read();
	if (s == "")
	    break;
	f->write(s);
	if (dump_commands_file)
	    Stdio.append_file(dump_commands_file, s);
    }
    f->close();
    if (debug)
	trace("Writer closed down\n");
}

// Start and/or connect to a GTP engine.
//
// This constructor has an admittedly awkward interface, but here goes:
//
// If program_start_array is non-empty, it is sent to
// Process.create_process() to start a new process. If also connection
// is empty, pipe communication is established with the new process.
//
// If connection contains one element, this is assumed to be a port to
// listen for a TCP connection on. That connection may come from the
// just started process, if any, or from somewhere else.
//
// If connection contains two elements, this is assumed to be a host
// and a port to make a TCP connection to. The listener may be the
// just started process, if any, or something else.
//
// If both program_start_array and connection are empty, no
// communication can possibly be established.
//
// Setting the debug_ argument lets you monitor the GTP communication.
// Setting crash_callback will give you a notice if the GTP
// communication is unexpectedly broken for some reason.
void create(array(string) program_start_array, array(string) connection,
	    int|void debug_, function|void crash_callback_)
{
    debug = debug_;
    crash_callback = crash_callback_;
    write_queue = Thread.Queue();
    object f1 = Stdio.File();
    object f2 = Stdio.FILE();
    object f3 = Stdio.FILE();
    mapping(string:object) pipes = ([]);
    Stdio.Port port;

    if (sizeof(program_start_array) == 0 && sizeof(connection) == 0)
    {
	werror("Pipe communication requested but no program started.\n");
	exit(1);
    }
    
    if (sizeof(connection) == 0)
    {
	// Communicate with pipes.
	pipes = (["stdin"  : f1->pipe(),
		  "stdout" : f2->pipe(),
		  "stderr" : f3->pipe()]);
    }
    else if (sizeof(connection) == 1)
    {
	// Listen for an incoming TCP connection. This will possibly
	// be made by the program we are about to start, so we must
	// start listening before doing that.
	int port_number = (int) connection[0];
        port = Stdio.Port();
	if (!port->bind(port_number))
	{
	    werror("Could not bind port %d.\n", port_number);
	    exit(1);
	}
    }
    
    if (sizeof(program_start_array) > 0)
	engine = Process.create_process(program_start_array, pipes);

    if (sizeof(connection) == 1)
    {
	// Listen for incoming TCP connection.
	f1 = port->accept();
	f2->assign(f1);
    }
    else if (sizeof(connection) == 2)
    {
	// Make an outgoing TCP connection. If we just started a
	// program that may listen, we should give it some time to
	// bind its port first.
	if (sizeof(program_start_array) > 0)
	    sleep(1);

	string host = connection[0];
	int port_number = (int) connection[1];
	if (!f1->connect(host, port_number))
	{
	    werror("Could not connect to %s %d.\n", host, port_number);
	    exit(1);
	}
	f2->assign(f1);
    }

    if (sizeof(connection) == 0)
	thread_create(program_trace_reader, f3);
    
    thread_create(program_reader, f2);
    thread_create(program_writer, f1);
}

void set_dump_commands_file(string filename)
{
    dump_commands_file = filename;
}

void set_dump_responses_file(string filename)
{
    dump_responses_file = filename;
}

void shutdown()
{
    quit();
    if (debug)
	trace("Program shut down.\n");
    // It might happen that the opponent adjourns a game because the
    // program is thinking for too long over a move. When that move
    // finally arrives, we may have started a new game, which would
    // be badly confused by this move. Thus we must remove all
    // callbacks now.
    callbacks = ([]);
    write_queue->write("");
    if (engine)
	destruct(engine);
}

void destroy()
{
    if (debug)
	trace("Program destroyed.\n");
}
