#!/usr/bin/env pike

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is gnugoclient, a program to connect go engines to go        *
 * servers. See http://www.lysator.liu.se/~gunnar/gnugoclient        *
 * for information.                                                  *
 *                                                                   *
 * Copyright 1998-2004 by Gunnar Farnebck.                          *
 * E-mail: gunnar@lysator.liu.se                                     *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License as    *
 * published by the Free Software Foundation - version 2             *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License in file COPYING for more details.      *
 *                                                                   *
 * You should have received a copy of the GNU General Public         *
 * License along with this program; if not, write to the Free        *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,           *
 * Boston, MA 02111, USA.                                            *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

string gnugoclient_version = "3.0";

// All available options are listed here together with default values.
mapping(string:string) options = ([
    "minboard" : "9",
    "maxboard" : "19",
    "keepalive" : "50",    // interval in minutes
    "server_address" : "",
    "server_port" : "9696",
    "server_protocol" : "nngs",
    "server_name" : "",    // used in PC property of sgf
    "command_line" : "",
    "gtp_connection_method" : "pipes",
    "gtp_host" : "",
    "gtp_port" : "",
    "blacklist" : "",
    "whitelist" : "",
    "min_main_time" : "0",
    "min_byo_time": "0",
    "name" : "",
    "password" : "",
    "sgf_copyright" : "",   // used in CP property of sgf
    "sgf_save_time_info" : "off",
    "interactive" : "on",
    "stone_removal_policy" : "automatic",
    "debug_gtp" : "off",
    "showboard_after_move" : "off",
    "supervisor" : "",     // Whom to refer problems with the program to.
    "logfile" : "./%s/log",
    "dump_gtp_commands" : "",
    "dump_gtp_responses" : "",
    "sgfname_white" : "./%s/%n%v-%o-%d.sgf",
    "sgfname_black" : "./%s/%o-%n%v-%d.sgf",
]);

// Those options which have a limited number of valid values are listed here.
mapping(string:array(string)) valid_option_values = ([
    "server_protocol" : ({"nngs", "pgs"}),
    "gtp_connection_method" : ({"pipes", "tcp_connect", "tcp_listen"}),
    "sgf_save_time_info" : ({"on", "off"}),
    "interactive" : ({"on", "off"}),
    "stone_removal_policy" : ({"automatic", "opponent", "manual"}),
    "debug_gtp" : ({"on", "off"}),
    "showboard_after_move" : ({"on", "off"})
]);


Stdio.FILE socket;
program Gtp = (program)"Gtp.pike";
object engine;
array(string) program_start_array = ({});
string logfile = "";
int engine_active = 0;

string current_opponent = "";
mapping(string:int) desired_handicap = ([]);
string playing = "free";
string opponent_color = "W";
int handicap = 0;
int board_size;
string engine_name;
string engine_version;
string gtp_protocol_version;
array(string) known_commands;
string last_move = "";
string myname = "";
int move_number = 0;
int start_at_move = 0;
int scoring = 0;
mapping(string:string) stored_games = ([]);
string pending_command = ""; // Ugly hack.
int time_to_reconnect = 1;
int do_reconnect = 1;
int manual_stone_removal = 0; // Remove dead stones by hand. Only relevant when
                              // stone_removal_policy is set to opponent.
                              // Temporarily set when manually issuing a match
                              // command.
.Sgf.SgfNode sgfroot;
.Sgf.SgfNode sgf;
string sgf_filename;
mapping(string:int) dead_stones = 0;
array(string) removed_stones = ({});
int complaints = 0;

int current_main_time;
int current_byo_yomi_time;

// Write a message to the user. When running in screen, a CR seems to
// be required.
void trace(string s)
{
    if (options->interactive == "on")
	werror(replace(s, "\n", "\r\n"));
}

void tell_server(string s)
{
    trace("Tell server: " + s + "\n");
    socket->write(s + "\n");
}

void tell_opponent(string s)
{
    tell_server("tell " + current_opponent+ " " + s);
}

// Send the showboard command to the engine if the showboard_after_move
// option is turned on and the engine supports the command. Only print
// the result if the debug_gtp option is off.
void do_showboard()
{
    if (options->showboard_after_move == "on"
	&& has_value(known_commands, "showboard"))
    {
	string board = engine->send_command_sync("showboard");
	if (options->debug_gtp == "off")
	    write(replace(board + "\n", "\n", "\r\n"));
    }
}

void relay_move(string move)
{
    last_move = move;
    tell_server(move);
    if (move != "resign")
    {
	sgf = .Sgf.SgfNode(sgf);
	if (opponent_color == "W")
	    sgf->add_move("B", move, board_size);
	else
	    sgf->add_move("W", move, board_size);
	save_sgf_file();
    }

    thread_create(do_showboard);
}

void store_handicap_stones(string stones)
{
    if (!sgfroot)
	return;
    sgfroot->remove_property("AB");
    foreach(stones / " ", string vertex)
	sgfroot->add_stone("B", vertex, board_size);
    save_sgf_file();
}

void crash_callback()
{
    add_sgf_comment("Program crashed.");
    tell_opponent("I'm in trouble, the program crashed.");
    log_event("Program crashed.");
    tell_server("quit");
}

array(string) process_stone_list(string stones, string status)
{
    array(string) groups = ({});
    if (status == "dead")
	dead_stones = ([]);
    foreach (stones / "\n", string line)
    {
	if (sizeof(line) == 0)
	    continue;
	groups += ({ (line / " ")[0] });
	if (status == "dead")
	    foreach (line / " ", string stone)
		dead_stones[stone] = 1;
    }
    return groups;
}

void remove_stones()
{
    string dead = engine->send_command_sync("final_status_list dead");
    array(string) dead_groups = process_stone_list(dead, "dead");
    if (sizeof(dead_groups) == 0)
	tell_opponent("All stones seem to be alive.");
    else if (sizeof(dead_groups) == 1)
	tell_opponent("I think " + dead_groups[0] + " is dead.");
    else
	tell_opponent("I think " + (dead_groups * ", ") + " are dead.");
    
    string seki = engine->send_command_sync("final_status_list seki");
    array(string) seki_groups = process_stone_list(seki, "seki");
    if (sizeof(seki_groups) > 0)
	tell_opponent(seki_groups * ", " + " live in seki.");

    foreach (dead_groups, string stone)
	tell_server(stone);

    complaints = 0;
    foreach (removed_stones, string stone)
    {
	if (!dead_stones[stone])
	{
	    complain_about_removed_stone(stone);
	    complaints++;
	}
    }
    if (complaints == 0)
	tell_server("done");
}

void final_score_callback(string s)
{
    add_sgf_comment("final_score: " + s);
    // It's vital to give remove_stones a thread on it's own since it does 
    // a send_command_sync, which deadlocks if it's being executed by
    // the callback thread.
    thread_create(remove_stones);
}

void complain_about_removed_stone(string where)
{
    tell_opponent("I don't agree about " + where + " being dead.");
    tell_opponent("Please adjourn and restart the scoring if you removed it by mistake.");
    if (options->supervisor != "")
	tell_opponent("If you think I'm wrong, please adjourn and ask "
		      + options->supervisor + " to investigate the problem.");
    tell_server("adjourn");
}

string date()
{
    mapping(string:int) t = localtime(time());
    return sprintf("%d%02d%02d%02d%02d", 1900 + (t->year), 1 + (t->mon),
		   t->mday, t->hour, t->min);
}

string filename_substitutions(string s, int|void logfile)
{
    s = replace(s, "%s", options->server_name);
    s = replace(s, "%N", options->name);

    if (logfile)
    {
	if (has_value(s, "%n") || has_value(s, "%v") || has_value(s, "%o"))
	    error("logfile cannot contain %n, %v, or %o codes\n");
    }
    else
    {
	s = replace(s, "%n", engine_name);
	s = replace(s, "%v", engine_version);
	s = replace(s, "%o", current_opponent);
    }
    
    return s;
}

void fix_directories(string filename)
{
    if (filename == "")
	return;
    array(string) path = explode_path(Stdio.simplify_path(filename));
    if (sizeof(path) > 1)
    {
	string dirname = combine_path(@path[0..sizeof(path)-2]);
	if (!Stdio.mkdirhier(dirname))
	    error("Could not create directories for " + filename + "\n");
    }
}

// Generate the sgf filename to use for this game.
string new_sgf_filename()
{
    string this_date = date();

    if (opponent_color == "W")
	sgf_filename = options->sgfname_white;
    else
	sgf_filename = options->sgfname_black;

    sgf_filename = filename_substitutions(sgf_filename);
    sgf_filename = replace(sgf_filename, "%d", this_date);
    fix_directories(sgf_filename);
}

void start_game()
{
    string color = "W";
    if (opponent_color == "W")
	color = "B";
    move_number = 0;
    scoring = 0;
    new_sgf_filename();
    engine_active = 1;
    // FIXME: mechanism to set random seed.
}

void start_sgf_tree(string white_player, string white_rank,
		    string black_player, string black_rank,
		    int boardsize, string komi, int handicap)
{
    if (white_player == myname)
	white_player += sprintf(" (%s %s)", engine_name, engine_version);

    if (black_player == myname)
	black_player += sprintf(" (%s %s)", engine_name, engine_version);

    sgfroot = .Sgf.SgfNode();
    sgf = sgfroot;
    sgfroot->add_property("GM", "1");
    sgfroot->add_property("FF", "4");
    sgfroot->add_property("AP", "gnugoclient:" + gnugoclient_version);
    if (options->sgf_copyright != "")
	sgfroot->add_property("CP", options->sgf_copyright);
    mapping(string:int) m = localtime(time());
    sgfroot->add_property("DT", sprintf("%4d-%02d-%02d", 1900 + m["year"],
				       m["mon"] + 1, m["mday"]));
    if (options->server_name != "")
	sgfroot->add_property("PC", options->server_name);
    sgfroot->add_property("PW", white_player);
    sgfroot->add_property("WR", white_rank);
    sgfroot->add_property("PB", black_player);
    sgfroot->add_property("BR", black_rank);
    sgfroot->add_property("SZ", (string) boardsize);
    sgfroot->add_property("KM", komi);
    sgfroot->add_property("HA", (string) handicap);
    save_sgf_file();
}

void add_sgf_comment(string s)
{
    if (!sgf)
	return;
    sgf->add_property("C", s);
    save_sgf_file();
}

void add_root_property(string property, string value)
{
    if (!sgfroot)
	return;
    sgfroot->add_property(property, value);
    save_sgf_file();
}

void save_sgf_file()
{
    if (!sgfroot || sgf_filename == "")
	return;
    Stdio.write_file(sgf_filename, sgfroot->write() + "\n");
}

string log_date()
{
    mapping(string:int) t = localtime(time());
    return sprintf("%02d%02d%02d %02d:%02d:%02d", (t->year) % 100,
		   1 + (t->mon), t->mday, t->hour, t->min, t->sec);
}

void log_event(string s)
{
    if (logfile != "")
	Stdio.append_file(logfile, log_date() + " " + s + "\n");
}

// Overload this method and the next if you have more complex
// limitations on board size.
int acceptable_size(int n)
{
    if (n < (int) options->minboard
	|| n > (int) options->maxboard)
	return 0;
    
    string response = engine->send_command_sync("boardsize " + n);
    if (response != "")
	return 0;

    return 1;
}

string why_unacceptable_size(int n)
{
    int min = (int) options->minboard;
    int max = (int) options->maxboard;
    if (n < min)
	return sprintf("I don't play on smaller boards than %dx%d, sorry.",
		       min, min);
    if (n > max)
	return sprintf("I don't play on larger boards than %dx%d, sorry.",
		       max, max);

    return sprintf("I don't play on %dx%d, sorry.", n, n);
}

// Overload this method and the next if you have more complex
// time requirements.
int acceptable_time(int main_time, int byo_yomi_time)
{
    // Zero main time means no time control.
    if (main_time == 0)
	return 1;
    
    return (main_time >= (int) options->min_main_time
	    && byo_yomi_time >= (int) options->min_byo_time);
}

string why_unacceptable_time(int main_time, int byo_yomi_time)
{
    int min_main = (int) options->min_main_time;
    int min_byo = (int) options->min_byo_time;
    
    if (main_time < min_main)
	return sprintf("I need at least %d minutes main time.", min_main);
    if (byo_yomi_time < min_byo)
	return sprintf("I need at least %d minutes byo yomi time.", min_byo);
}

// Do we want to play against this opponent.
// If we have a whitelist, only accept opponents in the list.
// Otherwise accept the opponent unless included in a blacklist.
int acceptable_opponent(string opponent)
{
    if (options->whitelist != ""
	&& !has_value(options->whitelist, opponent))
	return 0;
    else if (has_value(options->blacklist, opponent))
	return 0;

    return 1;
}

// Send time reports to the engine.
void time_report(string color, int white_time, int white_moves,
		 int black_time, int black_moves)
{
    if (options->sgf_save_time_info == "on")
    {
	if (color == "W")
	{
	    sgf->add_property("WL", (string) white_time);
	    if (white_moves > 0)
		sgf->add_property("OW", (string) white_moves);
	}
	else
	{
	    sgf->add_property("BL", (string) black_time);
	    if (black_moves > 0)
		sgf->add_property("OB", (string) black_moves);
	}
	save_sgf_file();
    }
    
    if (!has_value(known_commands, "time_left"))
	return;
    
    if (color == "W")
	engine->send_command(sprintf("time_left white %d %d", white_time,
				     white_moves < 0 ? 0 : white_moves));
    
    if (color == "B")
	engine->send_command(sprintf("time_left black %d %d", black_time,
				     black_moves < 0 ? 0 : black_moves));
}

// Overload this method if you want to do something whenever a game
// has ended.
void game_finished()
{
}

void generate_move()
{
    string prefix = "genmove ";
    if (gtp_protocol_version == "1")
        prefix = "genmove_";

    if (opponent_color == "B")
        engine->send_command(prefix + "white", relay_move);
    else
        engine->send_command(prefix + "black", relay_move);
}

void play_move(string color, string move)
{
    sgf = .Sgf.SgfNode(sgf);
    string prefix = "play ";
    if (gtp_protocol_version == "1")
      prefix = "";

    if (color == "W")
    {
        engine->send_command(prefix + "white " + move);
	sgf->add_move("W", move, board_size);
    }
    else
    {
        engine->send_command(prefix + "black " + move);
	sgf->add_move("B", move, board_size);
    }
    save_sgf_file();

    do_showboard();
}

// Overload this if you want your program to take some number of
// handicap stones per default.
int default_handicap(int size)
{
    return 0;
}

void parse_server_message(int message_code, array(string) message)
{
    if (message_code == 15)
	parse_message_15(message);
    else
	foreach (message, string s)
	    parse_server_message_old(message_code + " " + s);
}

// Message containing time info and made move;
void parse_message_15(array(string) message)
{
//    werror("%O\n", message);
    if (sizeof(message) < 2)
	return;

    // Time report for a move.
    string white_player;
    string black_player;
    int white_time;
    int white_moves;
    int black_time;
    int black_moves;
    int pending_time_report = 0;
    if (sscanf(message[0], "Game %*d %*s: %s (%*d %d %d) vs %s (%*d %d %d)",
	       white_player, white_time, white_moves, black_player,
	       black_time, black_moves) == 10)
    {
	if (engine_active && sizeof(message) == 2)
	    pending_time_report = 1;
	else
	{
	    // If somebody has restarted an adjourned game, this is
	    // our only chance to even get partial information about
	    // the initial main time. We set it to the nearest full
	    // minute more than the highest currently remaining main
	    // time.
	    current_main_time = 1;
	    if (white_moves == -1
		&& (white_time - 1) / 60 + 1 > current_main_time)
		current_main_time = (white_time - 1) / 60 + 1;
	    if (black_moves == -1
		&& (black_time - 1) / 60 + 1 > current_main_time)
		current_main_time = (black_time - 1) / 60 + 1;
	}
    }
    
    foreach (message[1..], string s)
    {
	// A move was made.
	string move;
	int number;
	string color;
	if (sscanf(s, "%d(%s): %s", number, color, move) > 0)
	{
	    // Ignore moves appearing before the response of the games command.
	    // This typically happens when someone has restarted an
	    // adjourned game.
	    if (!engine_active)
		return;
	    
#if 0
		werror("opponent_color = " + opponent_color + "\r\n");
		werror("number = " + number + "\r\n");
		werror("move_number = " + move_number + "\r\n");
		werror("start_at_move = " + start_at_move + "\r\n");
#endif
		
	    move = lower_case(move);
	    if (number != move_number)
	    {
		// The first case should only happen if the handicap is
		// undone and then changed. Otherwise it is an undone move
		// or an early (before the "games" report) move causing
		// this, which we can safely ignore. 
		if (move[0..7] == "handicap")
		{
		    engine->send_command("clear_board");
		    engine->send_command("fixed_" + move,
					 store_handicap_stones);
		    add_root_property("HA", move - "handicap ");
		}
		else
		    trace("Ignored move.\n");
		continue;
	    }
	    
	    if (number < start_at_move)
	    {
		if (move[0..7] == "handicap")
		{
		    engine->send_command("clear_board");
		    engine->send_command("fixed_" + move,
					 store_handicap_stones);
		    add_root_property("HA", move - "handicap ");
		}
		else
		    play_move(color, move);
		
		move_number++;
		if (number == start_at_move - 1 && color == opponent_color)
		    generate_move();
		continue;
	    }
	    
	    if (color == opponent_color)
	    {
		if (sscanf(move, "handicap %d", handicap) > 0)
		{
		    engine->send_command("clear_board");
		    engine->send_command("fixed_" + move,
					 store_handicap_stones);
		    add_root_property("HA", move - "handicap ");
		    if (color == "B")
			generate_move();
		}
		else
		{
		    play_move(opponent_color, move);
		    if (pending_time_report)
		    {
			time_report(color, white_time, white_moves,
				    black_time, black_moves);
			pending_time_report = 0;
		    }
		    generate_move();
		}
	    }
	    
	    if (pending_time_report && !has_value(move, "handicap"))
		time_report(color, white_time, white_moves,
			    black_time, black_moves);
	    
	    move_number++;
	}
    }
}

void parse_server_message_old(string s)
{
    // We're being challenged.
    string match;
    if ((playing == "free") && (sscanf(s, "9 Use <match %s>", match) > 0))
    {
	string opponent, color;
	int size, time, byo;
	sscanf(match, "%s %s %d %d %d", opponent, color, size, time, byo);
	current_opponent = opponent;

	if (!acceptable_size(size))
	{
	    tell_server("decline " + opponent);
	    tell_opponent(why_unacceptable_size(size));
	    log_event("Declined game with " + opponent +
		      ". Unacceptable size " + size + ".");
	    return;
	}
	
	if (!acceptable_time(time, byo))
	{
	    tell_server("decline " + opponent);
	    tell_opponent(why_unacceptable_time(time, byo));
	    log_event("Declined game with " + opponent +
		      ". Unacceptable time " + time + "/" + byo + ".");
	    return;
	}
	
	if (stored_games[lower_case(opponent)])
	{
	    tell_opponent(" It seems we have a stored game that we could resume instead.");
	    tell_server("decline " + opponent);
	    log_event("Declined game with " + opponent +
		      ". Stored game exists.");
	    tell_server("load " + stored_games[lower_case(opponent)]);
	    return;
	}
	
	if (!acceptable_opponent(opponent))
	{
	    tell_server("decline " + opponent);
	    log_event("Declined game with " + opponent + ", blacklisted.");
	    return;
	}
	
	if (options->server_protocol == "pgs")
	    tell_server("accept " + opponent);
	else
	    tell_server("match " + match);
	manual_stone_removal = 0;
	log_event("Accepted game: " + opponent + " " + color + " " +
		  size + " " + time + " " + byo);
	current_main_time = time;
	current_byo_yomi_time = byo;
	
	handicap = default_handicap(size);
	
	if (color=="B")
	{
	    if (!zero_type(desired_handicap[opponent]))
		handicap = desired_handicap[opponent];
	    tell_server("handicap " + handicap);
	}
	else
	    opponent_color = "B";
	
	return;
    }

    // Game message
    if (s[0..1] == "7 " && s[2..5] != "[##]")
    {
	string white_player, white_rank, black_player, black_rank, komi;
	int this_start_at_move, this_board_size, this_handicap;
	sscanf(s, "7 [%*d]%*[ ]%s [%s] vs.%*[ ]%s [%s] (%d %d %d%*[ ]%s %d",
	       white_player, white_rank, black_player, black_rank,
	       this_start_at_move, this_board_size, this_handicap, komi,
	       current_byo_yomi_time);
	if (white_player == myname)
	{
	    opponent_color = "B";
	    current_opponent = black_player;
	}
	else if (black_player == myname)
	{
	    opponent_color = "W";
	    current_opponent = white_player;
	}
	else
	    return;

	start_at_move = this_start_at_move;
	board_size = this_board_size;
	handicap = this_handicap;

	start_game();
	engine->send_command("boardsize " + board_size);
	engine->send_command("clear_board");
	engine->send_command("komi " + komi);
	if (has_value(known_commands, "time_settings"))
	    engine->send_command(sprintf("time_settings %d %d 25",
					 60 * current_main_time,
					 60 * current_byo_yomi_time));
	start_sgf_tree(white_player, String.trim_whites(white_rank),
		       black_player, String.trim_whites(black_rank),
		       board_size, komi, handicap);
	if (start_at_move > 0)
	{
	    tell_server("moves");
	}
	else if (opponent_color == "W")
	    generate_move();
	return;
    }


    // Start removing stones.
    if (has_prefix(s, "9 You can check your score"))
    {
	if (options->stone_removal_policy == "automatic")
	{
	    dead_stones = 0;
	    removed_stones = ({});
	    tell_opponent("I will now try to decide the final status of the stones as robustly as possible. This might take a short while. When I'm ready I'll tell you my conclusion and do the removals.");
	    engine->send_command("final_score", final_score_callback);
	}
	if (options->server_protocol != "pgs"
	    && options->stone_removal_policy == "opponent"
	    && !manual_stone_removal)
	    tell_server("done");
	scoring = 1;
    }

    // The opponent has removed one or more stones.
    string where;
    if (sscanf(s, "9 Removing @ %s", where) == 1)
    {
	add_sgf_comment(where + " removed");
	if (options->stone_removal_policy == "automatic")
	{
	    if (!dead_stones)
		removed_stones += ({where});
	    else
	    {
		if (!dead_stones[where])
		{
		    complaints++;
		    complain_about_removed_stone(where);
		}
	    }
	}
	if (options->server_protocol != "pgs"
	    && options->stone_removal_policy == "opponent"
	    && !manual_stone_removal)
	    tell_server("done");
    }

    // The opponent removed stones that we had already removed.
    // (Probably, although I'm not at all certain.)
    if (sscanf(s, "9 %*s was not finished removing stones.  Please type \"done\" again%*s") == 2)
    {
	if (complaints == 0)
	    tell_server("done");
    }

    // This is only effective on PGS.
    if (s == "9 Your opponent has typed done.")
	if (options->stone_removal_policy == "opponent"
	    && !manual_stone_removal)
	    tell_server("done");

    // Not certain whether this is necessary.
    if (options->server_protocol != "pgs"
	&& (s[0..1] == "9 ")
	&& has_prefix(s, "Please type \"done\" again.")
	&& options->stone_removal_policy == "opponent"
	&& !manual_stone_removal)
	tell_server("done");

    // Score report.
    if (s[0..2] == "20 ")
	;
    
    // Someone telling us their preferred handicap.
    string who;
    int handi;
    // The latter syntax is used by the NNGS versions at sourceforge.
    if (sscanf(s, "24 *%s*:%*[ ]handi%*[^ ]%*[ ]%d", who, handi) == 5
	|| sscanf(s, "24 %s -->%*[ ]handi%*[^ ]%*[ ]%d", who, handi) == 5)
    {
	if (handi <= 1)
	    handi = 0;
	if (handi > 9)
	    handi = 9;
	desired_handicap[who] = handi;
	tell_server("tell " + who + " Desired handicap set to " + handi + ".");
	return;
    }
    else if (s[0..2] == "24 ")
    {
	log_event(s[3..]);   // Log all other tells.
	return;
    }

    // Someone talking to us in a game.
    if (s[0..2] == "19 ")
    {
	log_event(s[3..]);
	return;
    }

    // Someone has restarted a game.
    string opponent;
    if (sscanf(s, "9 %s has restarted your game.", opponent) > 0)
    {
	tell_server("game " + myname);
	log_event("Game against " + opponent + " restarted.");
	return;
    }

    // Adjourn is requested, accept.
    if (options->server_protocol == "pgs")
    {
	if (s[0..1] == "9 " && has_value(s, "Use \"adjourn\""))
	{
	    tell_server("adjourn");
	    log_event("Adjourn requested.");
	    return;
	}
    }
    else
    {
	if (s[0..1] == "9 " && has_value(s, "like to adjourn"))
	{
	    tell_server("adjourn");
	    log_event("Adjourn requested.");
	    return;
	}
    }

    // Report on stored games.
    if (s[0..2] == "23 ")
    {
	foreach (((s - "23")/" ") - ({}), string stored)
	{
	    foreach (stored / "-", string player)
		stored_games[player] = stored;
	}
    }

    // A game we are playing has ended.
    string white, black, reason;
    if (sscanf(s, "9 {Game %*d: %s vs %s : %s", white, black, reason) == 4)
    {
	string who;
	if (white == myname)
	    who = black;
	else
	    who = white;
	who = lower_case(who);
	if (has_prefix(reason, "adjourned")
	    && move_number > 12
	    && board_size == 19)
	    stored_games[who] = white + "-" + black;
	else
	    m_delete(stored_games, who);
	log_event(white + " vs " + black + " (at move " + move_number +
		  ") : " + reason - "}");

	string sgf_result = "";
	string loser;
	float white_score, black_score;
	if (sscanf(reason, "%s resigns. W %f B %f",
		   loser, white_score, black_score) == 3)
	{
	    string winner = (white_score > black_score) ? "W" : "B";
	    sgf_result = sprintf("%s+%.1f", winner,
				 abs(white_score - black_score));
	}
	else if (sscanf(reason, "%s resigns.", loser) == 1)
	    sgf_result = sprintf("%s+Resign", (loser == "White") ? "B" : "W");
	else if (sscanf(reason, "%s forfeits", loser) == 1)
	    sgf_result = sprintf("%s+Time", (loser == "White") ? "B" : "W");

	if (sgf_result != "")
	    add_root_property("RE", sgf_result);

	if (has_prefix(reason, "adjourned"))
	{
	    if (engine)
		engine->remove_callback(relay_move);
	    add_sgf_comment("game adjourned");
	}

    }
    
    // Undo message.
    if (s[0..2] == "28 ") // undo
    {
	string undid_move;
	sscanf(s, "28 %*s undid the last move (%s)", undid_move);
	trace("Undo detected.\n");
	// Just replay the move.
	tell_server(undid_move);
	return;
    }

    // We have messed up.
    if (s == "5 Illegal Move!")
    {
	add_sgf_comment("Oops, that move was illegal.");
	tell_server("resign");
	log_event("!!!! Illegal move (" + last_move + ")");
	return;
    }

    // Opponent wants to change komi.
    // Only accept before or at move 5 so the opponent can't suddenly
    // adjust when he finds how much is needed to win. To avoid
    // attempts to trick the server's rating system we only accept
    // small komis. (E.g. nine handicap stones and komi -99 would
    // equate to about 18 stones in NNGS' rating system.)
    string komistring;
    float komi;
    if (sscanf(s, "9 Use \"komi %s\" to accept", komistring) == 1)
    {
	komi = (float) komistring;
	if (move_number <= 5 && komi > -11 && komi < 11)
	{
	    tell_server("komi " + komi);
	    if (engine)
	    {
		engine->send_command("komi " + komistring);
		add_root_property("KM", komistring);
	    }
	}
	else
	{
	    tell_server("decline komi");
	}
    }

    // The server informs us about the komi.
    if (sscanf(s, "9 Set the komi to %s", komistring) == 1
	|| sscanf(s, "9 Komi is now set to %s", komistring) == 1)
    {
	if (engine)
	{
	    engine->send_command("komi " + komistring);
	    add_root_property("KM", komistring);
	}
    }
    
    // The game has ended.
    if (playing == "playing" && s == "1 5")
    {
	engine_active = 0;
	if (sgfroot) {
	    save_sgf_file();
	    sgfroot = 0;
	    sgf = 0;
	}
	game_finished();
	
	playing = "free";
	return;
    }

    // A game has started.
    if (playing != "playing" && s == "1 6")
    {
	playing = "playing";
	dead_stones = 0;
	removed_stones = ({});
	return;
    }

    // A game has been created.
    if (has_prefix(s, "9 Creating match"))
    {
	if (pending_command != "")
	{
	    sleep(1);
	    tell_server(pending_command);
	    pending_command = "";
	}
	tell_server("game " + myname);
	return;
    }
}

void server_listener()
{
    int is_file = 0;
    int message_code = -1;
    array(string) message = ({});

    while (1)
    {
	string s = socket->gets();
	if (!s)
	    break;
	s -= "\r";
	trace("Server: " + s + "\n");

	int code;
	string text;

	if (is_file)
	{
	    if (s != message_code + " File")
	        message += ({s});
	    else
	    {
	        parse_server_message(message_code, message);
	        is_file = 0;
		message_code = -1;
		message = ({});
	    }
	}
	else if (sscanf(s, "%d %s", code, text) == 2)
	{
	    if (code != message_code && message_code != -1)
	    {
	        parse_server_message(message_code, message);
		message = ({});
		message_code = -1;
	    }

	    if (message_code == -1 && text == "File")
	    {
		is_file = 1;
		message_code = code;
	    }
	    else
	    {
		message_code = code;
		message += ({text});
	    }

	    // Immediately relay a prompt.
	    if (message_code == 1)
	    {
	        parse_server_message(message_code, message);
		message = ({});
		message_code = -1;
	    }
	}
	time_to_reconnect = 1;
    }
    trace("Lost contact to server.\n");
    log_event("Lost contact to server.\n");
    if (engine_active)
    {
	engine->remove_callback(relay_move);
	engine_active = 0;
	game_finished();
    }
    if (sgfroot) {
	save_sgf_file();
	sgfroot = 0;
	sgf = 0;
    }
    reconnect_to_server();
}

void login()
{
    tell_server(myname);
    if (options->password != "")
	tell_server(options->password);
    tell_server("set client on");
    tell_server("set robot 1");
    thread_create(update_stored_games);
    log_event("---------------- Logged in. ----------------");
    string my_realname = engine_name + " Version " + engine_version;
    tell_server("set realname " + my_realname);
}

// When stored games become too old, the server will silently purge them.
// Thus we update the list every 6 hours in order not to deny a game to
// somebody who can no longer load the old game.
void update_stored_games()
{
    while (1)
    {
	stored_games = ([]);
	tell_server("stored");
	sleep(6 * 60 * 60);
    }
}

void reconnect_to_server()
{
    if (!do_reconnect)
	return; // Don't reconnect.
    sleep(time_to_reconnect);
    trace("Trying to reconnect.");
    if (time_to_reconnect < 600) // 20 minutes
	time_to_reconnect *= 2;
    socket->close();
    socket->connect(options->server_address, (int) options->server_port);
    thread_create(server_listener);
    login();
}

void connect_to_server()
{
    socket = Stdio.FILE();
    socket->open_socket();
    socket->connect(options->server_address, (int) options->server_port);
    thread_create(server_listener);
    login();
}

// On recent versions of the NNGS server code, this should not be
// necessary anymore.
void keep_alive(int interval)
{
    while (1)
    {
	sleep(interval * 60);
	tell_server("ayt");
    }
}

void exit_with_error(string s)
{
    trace(s);
    tell_server("quit");
    sleep(5);
    exit(1);
}    

void error(string s)
{
    werror(s);
    exit(1);
}

void process_option(string option, string|void filename)
{
    option = String.trim_whites(option);
    if (option == "")
	return;

    string name, value;
    if (sscanf(option, "%s=%s", name, value) != 2)
	error(sprintf("Bad option '%s'%s\n", option,
		      filename ? " in file " + filename : ""));
    name = String.trim_whites(name);
    value = String.trim_whites(value);

    if (!options[name])
	error(sprintf("Unknown option %s\n", name));

    array(string) alternatives = valid_option_values[name];
    if (alternatives && !has_value(alternatives, value))
	error(sprintf("Bad value %s for option %s.\nValid values are %s.\n",
		      value, name, alternatives * ", "));

    options[name] = value;
}

// Read in options from the command line and from indicated files.
void get_options(array(string) args)
{
    array(string) option_args = ({});
    foreach (args, string arg)
    {
	if (has_prefix(arg, "--"))
	    option_args += ({arg});
	else
	{
	    string s = Stdio.read_file(arg);
	    if (!s)
		error(sprintf("Could not read %s\n", arg));
	    
	    foreach (s / "\n", string row)
		process_option(row, arg);
	}
    }
    foreach (option_args, string option)
	process_option(option[2..]);
}

// Check the sanity of various settings and set up some derived variables.
void check_options()
{
    if (options->server_address == "")
	error("No server address specified.\n");

    if ((int) options->server_port == 0)
	error("No server port specified or not an integer.\n");

    if (options->command_line == "" && options->gtp_connection_method == "pipes")
	error("No command line given to start the program. Use --command_line option.\n");
    program_start_array = options->command_line / " " - ({""});

    if (options->gtp_connection_method == "tcp_listen" && options->gtp_port == "")
	error("No port specified but connection method is tcp_listen.\n");

    if (options->gtp_connection_method == "tcp_connect"
	&& (options->gtp_port == ""
	    || options->gtp_host == ""))
	error("No host or port specified but connection method is tcp_connect.\n");

    if (options->name == "")
	error("No name to use on server given.\n");
    myname = options->name;

    logfile = filename_substitutions(options->logfile, 1);
    fix_directories(logfile);
}

// Start the engine and/or connect to it, then get information about
// name and version.
void start_engine()
{
    array(string) connection = ({});

    if (options->gtp_connection_method == "tcp_listen")
	connection = ({options->gtp_port});
    else if (options->gtp_connection_method == "tcp_connect")
	connection = ({options->gtp_host, options->gtp_port});

    engine = Gtp(program_start_array, connection, options->debug_gtp == "on",
		 crash_callback);

    if (options->dump_gtp_commands != "")
	engine->set_dump_commands_file(filename_substitutions(
					   options->dump_gtp_commands, 1));
    if (options->dump_gtp_responses != "")
	engine->set_dump_responses_file(filename_substitutions(
					    options->dump_gtp_responses, 1));
    
    engine_name = engine->send_command_sync("name");
    engine_version = engine->send_command_sync("version");
    gtp_protocol_version = engine->send_command_sync("protocol_version");
    known_commands = engine->send_command_sync("list_commands") / "\n";
    log_event(sprintf("GTP engine is %s %s", engine_name, engine_version));

    if (options->stone_removal_policy == "automatic"
	&& !has_value(known_commands, "final_status_list"))
    {
	werror("Automatic stone removal cannot be used if the engine does\n"
	       + "not support the final_status_list command.\n"
	       + "Use --stone_removal_policy=opponent or\n"
	       + "--stone_removal_policy=manual instead.\n");
	exit(1);
    }
}

int main(int argc, array(string) argv)
{
    // Analyze the command line arguments.
    get_options(argv[1..]);
    check_options();

    // Now we can start the engine and connect to the server.
    start_engine();
    connect_to_server();

    // For some servers we may need to send commands regularly in order
    // not to be kicked off due to inactivity.
    if ((int) options->keepalive > 0)
	thread_create(keep_alive, (int) options->keepalive);

    // If interactivity is disabled, return a negative number
    // to hang around in the interpreter backend.
    if (options->interactive != "on")
	return -1;

    // Otherwise we start a readline loop.
    object readline = Stdio.Readline();
    readline->set_prompt("> ");

    while (1)
    {
	string s = readline->read();
	int use_handicap = 0;
	if (has_prefix(s, "match")) {
	    sscanf(s, "match %*s %*d %d %d",
		   current_main_time, current_byo_yomi_time);
	    manual_stone_removal = 1;
	    if (use_handicap != 0)
		pending_command = "handicap " + use_handicap;
	}
	else if (s == "quit")
	    do_reconnect = 0;

	if (has_prefix(s, "use_handicap"))
	    use_handicap = (int) (s - "use_handicap ");
	else
	    tell_server(s);

	if (has_prefix(s, "load"))
	    tell_server("game " + myname);
	    
	if (s == "quit")
	    break;
    }
    engine->shutdown();
    
    // Give the reader and the writer some time to shut down.
    sleep(5);
}
